/- Lecture 4.3: Mathematics — Rational and Real Numbers -/

import analysis.real

set_option pp.beta true


/- Rational numbers -/

namespace my_rat

structure pre :=
(num           : ℤ)
(denum         : ℤ)
(denum_ne_zero : denum ≠ 0)

namespace pre

instance : setoid pre :=
{ r     := λa b, a.num * b.denum = b.num * a.denum,
  iseqv :=
  ⟨ assume ⟨n, d, hd⟩, by simp,
    assume a b h, h.symm,
    assume a b c eq_ab eq_bc,
      have (a.num * c.denum) * b.denum = (c.num * a.denum) * b.denum :=
        calc (a.num * c.denum) * b.denum = (a.num * b.denum) * c.denum : by ac_refl
        ... = a.denum * (b.num * c.denum) : by rw [eq_ab]; ac_refl
        ... = (c.num * a.denum) * b.denum : by rw [eq_bc]; ac_refl,
      calc a.num * c.denum = c.num * a.denum :
        eq_of_mul_eq_mul_right b.denum_ne_zero this ⟩ }

@[simp] lemma setoid_iff (a b : pre) : a ≈ b ↔ a.num * b.denum = b.num * a.denum :=
iff.rfl

attribute [simp] pre.denum_ne_zero

def of_int (z : ℤ) : pre := ⟨z, 1, zero_ne_one.symm⟩

instance : has_zero pre := ⟨of_int 0⟩

@[simp] lemma zero_num : (0 : pre).num = 0 := rfl
@[simp] lemma zero_denum : (0 : pre).denum = 1 := rfl

instance : has_one pre := ⟨of_int 1⟩

instance : has_add pre :=
⟨λa b,
  { num           := a.num * b.denum + b.num * a.denum,
    denum         := a.denum * b.denum,
    denum_ne_zero := mul_ne_zero a.denum_ne_zero b.denum_ne_zero }⟩

@[simp] lemma add_num (a b : pre) : (a + b).num = a.num * b.denum + b.num * a.denum := rfl
@[simp] lemma add_denum (a b : pre) : (a + b).denum = a.denum * b.denum := rfl

lemma setoid_add {a b c d : pre} (hac : a ≈ c) (hbd : b ≈ d) : a + b ≈ c + d :=
begin
  simp at hac hbd,
  have h₁ : a.num * b.denum * (c.denum * d.denum) = (c.num * a.denum) * b.denum * d.denum :=
    by rw ← hac; ac_refl,
  have h₂ : b.num * a.denum * (c.denum * d.denum) = (d.num * b.denum) * a.denum * c.denum :=
    by rw ← hbd; ac_refl,
  simp [add_mul, h₁, h₂],
  ac_refl
end

instance : has_neg pre :=
⟨λa,
  { num           := - a.num,
    denum         := a.denum,
    denum_ne_zero := a.denum_ne_zero }⟩

@[simp] lemma neg_num (a : pre) : (- a).num = - a.num := rfl
@[simp] lemma neg_denum (a : pre) : (- a).denum = a.denum := rfl

lemma setoid_neg {a b : pre} (hac : a ≈ b) : - a ≈ - b :=
begin
  simp at hac ⊢,
  assumption
end

instance : has_mul pre :=
⟨λa b,
  { num           := a.num * b.num,
    denum         := a.denum * b.denum,
    denum_ne_zero := mul_ne_zero a.denum_ne_zero b.denum_ne_zero }⟩

@[simp] lemma mul_num (a b : pre) : (a * b).num = a.num * b.num := rfl
@[simp] lemma mul_denum (a b : pre) : (a * b).denum = a.denum * b.denum := rfl

lemma setoid_mul {a b c d : pre} (hac : a ≈ c) (hbd : b ≈ d) : a * b ≈ c * d :=
begin
  simp at hac hbd,
  have h : a.num * b.num * (c.denum * d.denum) = (a.num * c.denum) * (b.num * d.denum) :=
    by ac_refl,
  simp [h, hac, hbd],
  ac_refl
end

instance : has_inv pre :=
⟨λa, if ha : a.num = 0 then 0 else
  { num           := a.denum,
    denum         := a.num,
    denum_ne_zero := ha }⟩

lemma inv_def (a : pre) (ha : a.num ≠ 0) :
  a⁻¹ = { num := a.denum, denum := a.num, denum_ne_zero := ha } :=
dif_neg ha

@[simp] lemma inv_zero (a : pre) (ha : a.num = 0) : a⁻¹ = 0 :=
dif_pos ha

@[simp] lemma inv_num (a : pre) (ha : a.num ≠ 0) : (a⁻¹).num = a.denum :=
by rw [inv_def a ha]

@[simp] lemma inv_denum (a : pre) (ha : a.num ≠ 0) : (a⁻¹).denum = a.num :=
by rw [inv_def a ha]

lemma setoid_inv {a b : pre} (hac : a ≈ b) : a⁻¹ ≈ b⁻¹ :=
begin
  by_cases ha : a.num = 0; by_cases hb : b.num = 0; simp [ha, hb] at hac ⊢; try {contradiction},
  cc
end

def nonneg (a : pre) : Prop := 0 ≤ a.num * a.denum

lemma nonneg_iff (a : pre) :
  nonneg a ↔ (a.num = 0 ∨ (0 < a.num ∧ 0 < a.denum) ∨ (a.num < 0 ∧ a.denum < 0)) :=
match lt_trichotomy 0 a.num with
| or.inl ha          :=
  begin
    rw [nonneg, mul_comm, mul_nonneg_iff_right_nonneg_of_pos ha, le_iff_lt_or_eq, eq_comm],
    simp [ha, not_lt_of_gt ha, ne_of_gt ha]
  end
| or.inr (or.inl eq) := by simp [eq.symm, nonneg]
| or.inr (or.inr ha) :=
  begin
    simp [nonneg, ha, not_lt_of_gt ha, ne_of_lt ha],
    rw [mul_comm, ← neg_neg a.num, ← neg_neg a.denum, mul_neg_eq_neg_mul_symm, neg_mul_eq_neg_mul, neg_neg, neg_lt, neg_zero],
    rw [← neg_neg a.num, neg_lt, neg_zero] at ha,
    rw [mul_nonneg_iff_right_nonneg_of_pos ha, lt_neg, neg_zero, le_neg, neg_zero, le_iff_eq_or_lt],
    simp
  end
end

lemma setoid_nonneg.aux1 {a b : pre} (hab : a ≈ b) (hn : 0 < a.num) (hd : 0 ≤ a.denum) :
  nonneg b :=
begin
  have hd : 0 < a.denum := lt_of_le_of_ne hd a.denum_ne_zero.symm,
  have h : 0 ≤ a.num * b.denum ↔ 0 ≤ b.num * a.denum := by rw [(setoid_iff _ _).1 hab],
  rw [mul_comm, mul_nonneg_iff_right_nonneg_of_pos hn, mul_nonneg_iff_right_nonneg_of_pos hd] at h,
  by_cases hb : 0 ≤ b.num,
  { have hbd : 0 ≤ b.denum := by rwa [h],
    exact mul_nonneg hb hbd },
  { have hbd : ¬ 0 ≤ b.denum := by rwa [← h] at hb,
    exact mul_nonneg_of_nonpos_of_nonpos (le_of_not_ge hb) (le_of_not_ge hbd) }
end

lemma setoid_nonneg.aux2 {a b : pre} (hab : a ≈ b) (ha : nonneg a) : nonneg b :=
begin
  cases lt_trichotomy 0 a.num,
  { rw [nonneg, mul_comm, mul_nonneg_iff_right_nonneg_of_pos h] at ha,
    exact setoid_nonneg.aux1 hab h ha },
  cases h,
  { simp [h.symm] at hab,
    simp [nonneg, hab] },
  { have hnan : 0 < - a.num := lt_neg.2 h,
    have ha : 0 ≤ (- a.denum) * (- a.num) := by simp [nonneg, mul_comm] at ha ⊢; assumption,
    have hnad : 0 ≤ - a.denum := by rwa [mul_nonneg_iff_right_nonneg_of_pos hnan] at ha,
    let a' : pre := {
      num := - a.num,
      denum := - a.denum,
      denum_ne_zero := by rw [(≠), neg_eq_zero]; exact a.denum_ne_zero },
    have hab' : a' ≈ b := by simp at hab ⊢; assumption,
    exact setoid_nonneg.aux1 hab' hnan hnad }
end

lemma setoid_nonneg {a b : pre} (hab : a ≈ b) : nonneg a ↔ nonneg b :=
begin
  split,
  { exact setoid_nonneg.aux2 hab },
  { exact setoid_nonneg.aux2 hab.symm }
end

end pre

def rat : Type := quotient pre.setoid

instance : has_zero rat := ⟨⟦0⟧⟩

instance : has_one rat := ⟨⟦1⟧⟩

instance : has_add rat := ⟨ λa b, quotient.lift_on₂ a b (λa b, ⟦ a + b ⟧)
  begin
    assume a b c d hac hbd,
    apply quotient.sound,
    exact pre.setoid_add hac hbd
  end ⟩

instance : has_neg rat := ⟨ λa, quotient.lift_on a (λa, ⟦ - a ⟧)
  begin
    assume a b hab,
    apply quotient.sound,
    exact pre.setoid_neg hab
  end ⟩

instance : has_mul rat := ⟨ λa b, quotient.lift_on₂ a b (λa b, ⟦ a * b ⟧)
  begin
    assume a b c d hac hbd,
    apply quotient.sound,
    exact pre.setoid_mul hac hbd
  end ⟩

instance : has_inv rat := ⟨ λa, quotient.lift_on a (λa, ⟦ a⁻¹ ⟧)
  begin
    assume a b hab,
    apply quotient.sound,
    exact pre.setoid_inv hab
  end ⟩

def nonneg (r : rat) : Prop :=
quot.lift_on r pre.nonneg (assume a b hab, propext (pre.setoid_nonneg hab))

instance : has_le rat := ⟨ λa b, nonneg (b + - a) ⟩

end my_rat


/- Real numbers -/

#print real

section absolute_value

-- Absolute value `| | : β → ℚ` (**Note:** the definition is mathlib is more general)
variables (α : Type*) [ring α] (abs : α → ℚ) [is_absolute_value abs]

/- Definition of Cauchy sequences (parameteric in the absolute value) -/
lemma cau_seq_def : cau_seq α abs = {f : ℕ → α // is_cau_seq abs f} :=
by refl

lemma is_cau_seq_def (f : ℕ → α) :
  is_cau_seq abs f  ↔  ∀ ε > 0, ∃ i, ∀ j ≥ i, abs (f j - f i) < ε :=
iff.rfl

/- The quotient relation on Cauchy sequences -/
lemma equiv_def (f g : cau_seq α abs) :
  f ≈ g  ↔  ∀ ε > 0, ∃ i, ∀ j ≥ i, abs (f j - g j) < ε :=
iff.rfl

end absolute_value

/- Positive Cauchy sequences -/
lemma pos_def (f : cau_seq ℚ abs) :
  cau_seq.pos f  ↔  ∃ ε > 0, ∃ i, ∀ j ≥ i, ε ≤ f j :=
iff.rfl

/- Definition of ℝ -/
lemma real_def : real = @quotient (cau_seq ℚ abs) cau_seq.equiv :=
by unfold real; unfold cau_seq.completion.Cauchy
